# IoT Dashboard App

## Overview

![IOT Dashboard App screenshot](docs/iot-app-frontend.png)

This sample app consists of 3 parts.

Using the steps described here, an app that allows getting sensor data over MQTT, plotting the data
in POI details dialog and showing the ones that have exceeded an adjustable threshold can be developed.
Default POIs can be used alongside custom sensor POIs.

This example has 3 separate modules that operate individually: The frontend, MQTT broker and the backend.

Frontend is built using the NavVis IVION API. Backend uses Node.js and simulates the sensors. Both modules
are written with Typescript. MQTT broker is a Docker container with RabbitMQ MQTT Broker installed.

![IOT Dashboard App screenshot](docs/iot-app-architecture.png)


## 1. Backend

Backend is a simple Node.js app that simulates the sensors.

``TemperatureSensor`` and ``PressureSensor`` classes represents a sensor instance. They generate
pseudo-realistic sensory data that is generated within a range. When the data is generated, it is
published to the corresponding MQTT topic. (``/iot/temperature/{sensor_id}`` or ``iot/pressure/{sensor_id}``)

The backend application also exposes several endpoints. ``/temperature/{sensor_id}`` lists a temperature
sensor's historical data. ``/pressure/{sensor_id}`` lists a pressure sensor's historical data.

Also the sensor ids can be listed using ``/temperature/`` or ``/pressure/`` endpoints.

The backend can be run using the following commands in ``iot-dashboard-backend`` folder.

```bash
npm install
npm start
```

## 2. MQTT Broker

MQTT Broker is a standard Docker container with RabbitMQ MQTT broker installed.

Note that the broker needs to be configured to support MQTT over WebSockets.

Make sure ports 1883 and 15675 are accessible.

Once you create a working image you can run the following code to run it with needed ports.

```bash
sudo docker run -p 5672:5672 -p 15672:15672 -p 1883:1883 -p 15675:15675 -td IMAGE_NAME
```

> **Note:** This component isn't given in the sample app code, you need to create the Docker container yourself. 

##  3. Frontend

Main logic of the app resides in the frontend module. It takes care of getting sensor data over MQTT,
plotting the data in POI details dialog and showing the ones that have exceeded an adjustable threshold.

POIs that exceed a certain threshold will enter alert mode. Their icon will change and they will be
displayed in the sidebar menu dashboard.

This is accomplished by defining new POI groups and types. For example, in our case ``Temperature``
is the POI group and ``Temperature Normal`` and ``Temperature Alert`` are the types that belong to the group.

The POIs that have a sensor attached to them switch to the Alert POI type when the sensor they are
connected to exceeds a certain threshold value. They will return back to Normal POI type when the sensor
stops exceeding the threshold.

Default POIs can be used alongside custom sensor POIs.

What the frontend does can be summarized as follows:

1. Find all nearby POIs that have a sensor attached to them.
2. Subscribe to the MQTT topics of the nearby POIs.
3. Update the POIs according to the incoming data. (Write the sensor value to POI title, draw the chart, update POI type if necessary.)
4. Periodically update sidebar menu dashboard with POIs that have exceeded their threshold values. (Alert mode)

This app is organized as a bunch of reusable components.
Components will be described in their corresponding subsections.

It has 3 main components : ``PoiChart``, ``PoiConfiguration``, ``SidebarMenuDashboard``.
It also has several helper services : ``PoiHelperService``, ``SensorDataHelperService``, 
``ToggleablePoiService`` and ``CommunicationService``.

### 3.1 Steps to run the Frontend

While NavVis IVION is working on the background go to iot-dashboard-app folder and run the
following commands from the console.

```bash
npm install
npm run serve
```
### 3.2 Components

The app has 3 main components (``PoiChart``, ``PoiConfiguration``, ``SidebarMenuDashboard``) and
an abstract parent class ``PoiDialogModule``.

#### 3.2.1. PoiChart

![IOT Dashboard App POI Chart screenshot](docs/iot-dashboard-poi-chart.png)

``PoiChart`` is a class that contains all the functions that are needed to create and append a chart
to POI dialog. It extends the parent class ``PoiDialogModule``.

Its main function ``drawChart`` receives 3 parameters.

```typescript
// ./src/js/components/poi-chart/PoiChart.ts
...
public drawChart(chartData: FormattedSensorDatum[], threshold: number, yDomain: number[]): void
...
```

``chartData`` is the historical data of the current POI. It is the data that is listed by the backend
using ``/temperature/{sensor_id}`` or ``/pressure/{sensor_id}`` endpoints and formatted using
``SensorDataHelperService``.

``threshold`` is a number. It is used to draw the threshold line in the chart.

``yDomain`` is a number array with two elements. It is the range that chart will be drawn. For example,
if the yDomain is [10,100] the chart will have minimum value of 10 and maximum value of 100 on y axis.

Usage of this can be seen in the example below.

```typescript
// ./src/js/index.ts
...
this.poiChart.drawChart(this.historicalData, this.threshold, this.SENSOR_TYPES[type].range);
...
```
#### 3.2.2. PoiConfiguration

![IOT Dashboard App POI Configuration screenshot](docs/iot-dashboard-poi-configuration.png)

``PoiConfiguration`` is a class that appends a configuration panel that contains various input fields
to the POI dialog. It extends the parent class ``PoiDialogModule``.

Its main function ``addToDialog`` receives 1 parameter.

```typescript
// ./src/js/components/poi-configuration/PoiConfiguration.ts
...
public addToDialog(configs: Configuration): InputBoxesDictionary
...
```
``configs`` is the configuration template that is used to define input fields. It is of type Configuration.

```typescript
// ./src/js/components/poi-configuration/PoiConfiguration.ts
...
export type Configuration = {
  [key: string]: {
      label: string,
      type: ConfigsType,
      min?: number,
      max?: number,
      options?: HTMLOptionElement[],
      value: number
  };
}
...
```

``addToDialog`` function returns a dictionary of HTMLElement representing the input field boxes.
Key of each entry corresponds to the key of the ``Configuration`` object.

Usage of this can be seen in the example below.

```typescript
// ./src/js/index.ts
...
const configs: Configuration = {};
configs["threshold"] = {
    label: "Threshold",
    type: ConfigsType.SLIDER,
    min: this.SENSOR_TYPES[poiGroup].range[0],
    max: this.SENSOR_TYPES[poiGroup].range[1],
    value: customData.threshold || this.SENSOR_TYPES[poiGroup].threshold
};
configs["sensorId"] = {
    label: "Sensor ID",
    type: ConfigsType.DROPDOWN,
    options: options,
    value: customData.sensorId || 0
};
const inputBoxes = this.poiConfiguration.addToDialog(configs);
...
```

#### 3.2.3. SidebarMenuDashboard

![IOT Dashboard App POI Sidebar screenshot](docs/iot-dashboard-sidebar.png)

``SidebarMenuDashboard`` is a class that contains all the functions that are needed to create a
sidebar menu sensor dashboard. Please check ``Sidebar Menu`` guide to learn more about adding
custom sidebar menu items.

Its main function ``refreshItems`` receives 1 parameter.

```typescript
// ./src/js/components/sidebar-menu-dashboard/SidebarMenuDashboard.ts
...
public refreshItems(pois: PoiInterface[]): void
...
```
``pois`` is the PoiInterface array that will be shown in the sidebar menu.

```typescript
// ./src/js/index.ts
...
const poisOnAlert: PoiInterface[] = [];
for (const poi of Object.values(this.nearbyPois))
{
	const customData: PoiCustomData = JSON.parse(poi.customData);
	if (customData.value > customData.threshold)
	{
		poisOnAlert.push(poi);
	}
}
this.sidebarMenuDashboard.refreshItems(poisOnAlert);
...
```
### 3.3 Services

The app has 4 main services (``PoiHelperService``, ``SensorDataHelperService``, 
``ToggleablePoiService``, ``CommunicationService``).

#### 3.3.1. PoiHelperService

PoiHelperService is the service that contains helper functions to fetch or modify POIs.

It has 4 public functions.

``savePoi`` function saves the given POI to the server.

```typescript
// ./src/js/services/PoiHelperService.ts
/**
* Save the POI to the server.
* @param {PoiInterface} poi POI to be saved
* @returns {Promise<PoiInterface>} A promise with the saved POI object from the server.
*/
public savePoi(poi: PoiInterface): Promise<PoiInterface>
{
  return this.ivApi.poi.repository.save(poi).then((pois) =>
  {
    const savedPoi = pois[0];
    return savedPoi;
  });
}
```
``savePois`` function saves the given POI array to the server.

```typescript
// ./src/js/services/PoiHelperService.ts
/**
* Save the POIs to the server.
* @param {PoiInterface[]} pois POI to be saved
* @returns {Promise<PoiInterface>} A promise with the saved POI objects from the server.
*/
public savePois(pois: PoiInterface[]): Promise<PoiInterface[]>
{
  return this.ivApi.poi.repository.save(pois).then((pois) =>
  {
    const savedPois = pois;
    this.ivApi.poi.service.refreshPois();
    return savedPois;
  });
}
```
``updateCustomData`` function updates the customData of a specific POI. customData holds the
threshold value, sensor id and sensor value.

```typescript
// ./src/js/services/PoiHelperService.ts
/**
* Update the customData of a specific POI.
* @param {number} id Poi id
* @param {string} customData Poi custom data
* @returns {Promise<PoiInterface>} A promise with the saved POI object from the server.
*/
public updateCustomData(id: number, customData:string): Promise<PoiInterface>
{
  return this.ivApi.poi.repository.findOne(id).then((poi) =>
  {
    poi.customData = customData;
    return this.savePoi(poi).catch((e) => console.error(e));
  });
}
```

``initPoiTypes`` function checks whether required POI types and groups exist. 
If they don't, creates them. This is used to create necessary POI groups and types at the start of 
the app.

#### 3.3.2. SensorDataHelperService

SensorDataHelperService is the service that contains the functions needed to organize and format sensor data.

It has 3 functions.

``formatSensorData`` function formats the data taken from backend.

```typescript
// ./src/js/services/SensorDataHelperService.ts
/**
* Formats the sensor data, removes unnecessary fields, changes timestamp to a JS Date object.
* @param {SensorDatum[]} data Unformatted sensor data array
* @returns {FormattedSensorDatum[]} Formatted sensor data array
*/
public formatSensorData(data: SensorDatum[]): FormattedSensorDatum[]
	{
	const formattedData: FormattedSensorDatum[] = [];
	for (const datum of data)
	{
		formattedData.push({
			value: datum.value,
			date: new Date(datum.timestamp)
		});
	}
	return formattedData;
}
```
``getArrayWithLimitedLength`` function returns an empty array with a maximum length. If the array
tries to exceed that limit, the last element gets popped.

```typescript
// ./src/js/services/SensorDataHelperService.ts
/**
* Gives an array with a maximum length. If the array tries to exceed that limit, the last element gets popped.
* Used for storing sensor historical data.
* @param {number} length Maximum desired length of the array
* @returns {FormattedSensorDatum[]} Array with limited length
*/
public getArrayWithLimitedLength(length: number): FormattedSensorDatum[]
{
  const array: FormattedSensorDatum[] = [];
  array.push = function()
  {
    if (this.length >= length)
    {
      this.shift();
    }
    return Array.prototype.push.apply(this, arguments);
  };
  return array;
}
```

``isSensor`` function checks whether the given POI group supports adding sensors.

```typescript
// ./src/js/services/SensorDataHelperService.ts
/**
 * Checks whether the given POI group supports adding sensors.
 * @param {string} poiGroup POI group name
 * @param {SensorTypeDictionary} sensorTypes POI group name
 * @returns {boolean} true if POI group supports sensors
 */
public isSensor(poiGroup: string, sensorTypes: SensorTypeDictionary): boolean
{
	return (Object.keys(sensorTypes).includes(poiGroup));
}
```
#### 3.3.3. ToggleablePoiService

ToggleablePoiService is the service that contains all the functions that are needed to create a
toggleable POI. In the current system the POI type (Temperature, Pressure) is POI group type and
POI states (Alert, Normal) are POI types.

It has 2 functions.

``switchPoiToOn`` and ``switchPoiToOff`` works in similar ways.

``switchPoiToOn`` sets the POI type to the Alert state.

```typescript
// ./src/js/services/ToggleablePoiService.ts
public switchPoiToOn(poi: PoiInterface): PoiInterface
```
``switchPoiToOff`` sets the POI type to the normal type.

```typescript
// ./src/js/services/ToggleablePoiService.ts
 public switchPoiToOff(poi: PoiInterface): PoiInterface
```

#### 3.3.4. CommunicationService

ToggleablePoiService is the service that contains helper functions to fetch data 
from a remote location such as a server.

It has a main function called ``fetch`` and several functions that are dependent on the app.

``fetch`` is a private function that is used to send a HTTP GET request to a remote location.
It returns a promise in the type of the output data.
  
```typescript
	/**
	 * Fetch some data from an endpoint. Sends HTTP GET request, with response type = json.
	 * @param {string} endpoint POI to be saved
	 * @template T
	 * @returns {Promise<T>} A promise that holds the response of the GET request.
	 */
	private fetch<T>(endpoint: string): Promise<T>
```

Also ``CommunicationService`` has several app dependent public functions that uses ``fetch``.
Namely; ``fetchHistoricalData`` and ``fetchSensorIds``.

## 4. Used NavVis IVION Functionality

Many NavVis IVION API functionalities are used in order to build this app.

### 4.1 View API

View API is used to set the primary view of the application to 2D Map mode.

```typescript
this.ivApi.view.service.setPrimaryView(ViewType.MAP);
```
It is used to detect when the camera is moved to another location using the ``onTransitionEnd`` signal.

```typescript
this.ivApi.view.service.onTransitionEnd.connect(() =>
{
  this.subscribeToNearbyPois();
});
```
It is also used to get the ThreeJS camera instance to detect the distance between the camera and a POI.

```typescript
const camera = this.ivApi.view.mainView.getCamera();
```
### 4.2 UI API

UI API is used to add new sidebar menu items.

```typescript
this.ivApi.ui.sidebarMenuService.items.push(this.sidebarMenuDashboard.dashboard);
```
``closeMenu`` method is used to close the sidebar menu.

```typescript
this.ivApi.ui.sidebarMenuService.closeMenu();
```
### 4.3 POI API

POI API is used extensively throughout the application.

``onPoiOpen`` signal is used to check when a POI is clicked. It is used to append custom elements to
the default POI dialog.

``onPoiDelete`` is used to clean up the subscribed MQTT topics and several arrays after a POI delete.

``onPoiClose`` is used to clean up the current (open) POI variable.

``onPoiSave`` is used to assign default threshold and sensor id values if a newly created POI is in a
type that supports sensor data.

``findAll`` method is used to query all POIs.

```typescript
this.ivApi.poi.repository.findAll().then((pois) => { ... }
```
``findOne`` method is used to query one POI with its id.

```typescript
this.ivApi.poi.repository.findOne(id).then((poi) => { ... }
```
``unhighlightPois`` method is used to unhighlight given POIs.

```typescript
this.ivApi.poi.service.unhighlightPois(pois);
```
``highlightPois`` method is used to highlight given POIs.

```typescript
this.ivApi.poi.service.highlightPois(nearbyPois);
```
``goToPoi`` method is used to move the camera next to the POI.

```typescript
this.ivApi.poi.service.goToPoi(poi).catch((err) => {console.log(err)});
```
``openPoi`` method is used to open the POI dialog.

```typescript
this.ivApi.poi.service.openPoi(poi);
```
``save`` method is used to save a given POI or POI array to the server.

```typescript
return this.ivApi.poi.repository.save(poi).then((pois) =>
{
  const savedPoi = pois[0];
  return savedPoi;
});
```
### 4.4 Transform API

Transform API is used to get local location of a POI from its global location.

```typescript
let poiLocalLocation = this.ivApi.transform.service.globalToLocal.transform(poiGlobalLocation).clone();
```

### 4.5 Auth API

Auth API is used to get the current user.

```typescript
const user = this.ivApi.auth.currentUser;
```
